from .bind_file import BindFile, BindRecord, RecordClass
from .client import PKBClient, PKBClientException, API_ENDPOINT
from .dns import DNSRecord, DNSRestoreMode, DNSRecordType
from .domain import DomainInfo
from .forwarding import URLForwarding, URLForwardingType
from .ssl_cert import SSLCertBundle

__all__ = [
    "PKBClient",
    "PKBClientException",
    "API_ENDPOINT",
    "BindFile",
    "BindRecord",
    "RecordClass",
    "DNSRecord",
    "DNSRestoreMode",
    "DNSRecordType",
    "DomainInfo",
    "URLForwarding",
    "URLForwardingType",
    "SSLCertBundle",
]



import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

from pkb_client.client.dns import DNSRecordType, DNS_RECORDS_WITH_PRIORITY


class RecordClass(str, Enum):
    IN = "IN"

    def __str__(self):
        return self.value


@dataclass
class BindRecord:
    name: str
    ttl: int
    record_class: RecordClass
    record_type: DNSRecordType
    data: str
    prio: Optional[int] = None
    comment: Optional[str] = None

    def __str__(self):
        record_string = f"{self.name} {self.ttl} {self.record_class} {self.record_type}"
        if self.prio is not None:
            record_string += f" {self.prio}"
        record_string += f" {self.data}"
        if self.comment:
            record_string += f" ; {self.comment}"
        return record_string


class BindFile:
    origin: str
    ttl: Optional[int] = None
    records: List[BindRecord]

    def __init__(
        self,
        origin: str,
        ttl: Optional[int] = None,
        records: Optional[List[BindRecord]] = None,
    ) -> None:
        self.origin = origin
        self.ttl = ttl
        self.records = records or []

    @staticmethod
    def from_file(file_path: str) -> "BindFile":
        with open(file_path, "r") as f:
            file_data = f.readlines()

        # parse the file line by line
        origin = None
        ttl = None
        records = []
        for line in file_data:
            if line.startswith("$ORIGIN"):
                origin = line.split()[1]
            elif line.startswith("$TTL"):
                ttl = int(line.split()[1])
            else:
                # parse the records with the two possible formats:
                # 1: name 	ttl 	record-class 	record-type 	record-data
                # 2: name 	record-class 	ttl 	record-type 	record-data
                # whereby the ttl is optional

                # drop any right trailing comments
                line_parts = line.split(";", 1)
                line = line_parts[0].strip()
                comment = line_parts[1].strip() if len(line_parts) > 1 else None
                prio = None

                # skip empty lines
                if not line:
                    continue

                # find which format the line is
                record_parts = line.split()
                if record_parts[1].isdigit():
                    # scheme 1
                    if record_parts[3] not in DNSRecordType.__members__:
                        logging.warning(f"Ignoring unsupported record type: {line}")
                        continue
                    if record_parts[2] not in RecordClass.__members__:
                        logging.warning(f"Ignoring unsupported record class: {line}")
                        continue
                    record_name = record_parts[0]
                    record_ttl = int(record_parts[1])
                    record_class = RecordClass[record_parts[2]]
                    record_type = DNSRecordType[record_parts[3]]
                    if record_type in DNS_RECORDS_WITH_PRIORITY:
                        prio = int(record_parts[4])
                        record_data = " ".join(record_parts[5:])
                    else:
                        record_data = " ".join(record_parts[4:])
                elif record_parts[2].isdigit():
                    # scheme 2
                    if record_parts[3] not in DNSRecordType.__members__:
                        logging.warning(f"Ignoring unsupported record type: {line}")
                        continue
                    if record_parts[1] not in RecordClass.__members__:
                        logging.warning(f"Ignoring unsupported record class: {line}")
                        continue
                    record_name = record_parts[0]
                    record_ttl = int(record_parts[2])
                    record_class = RecordClass[record_parts[1]]
                    record_type = DNSRecordType[record_parts[3]]
                    if record_type in DNS_RECORDS_WITH_PRIORITY:
                        prio = int(record_parts[4])
                        record_data = " ".join(record_parts[5:])
                    else:
                        record_data = " ".join(record_parts[4:])
                else:
                    # no ttl, use default or previous
                    if record_parts[2] not in DNSRecordType.__members__:
                        logging.warning(f"Ignoring unsupported record type: {line}")
                        continue
                    if record_parts[1] not in RecordClass.__members__:
                        logging.warning(f"Ignoring unsupported record class: {line}")
                        continue
                    record_name = record_parts[0]
                    if ttl is None and not records:
                        raise ValueError("No TTL found in file")
                    record_ttl = ttl or records[-1].ttl
                    record_class = RecordClass[record_parts[1]]
                    record_type = DNSRecordType[record_parts[2]]
                    if record_type in DNS_RECORDS_WITH_PRIORITY:
                        prio = int(record_parts[3])
                        record_data = " ".join(record_parts[4:])
                    else:
                        record_data = " ".join(record_parts[3:])

                # replace @ in record name with origin
                record_name = record_name.replace("@", origin)

                records.append(
                    BindRecord(
                        record_name,
                        record_ttl,
                        record_class,
                        record_type,
                        record_data,
                        prio=prio,
                        comment=comment,
                    )
                )

        if origin is None:
            raise ValueError("No origin found in file")

        return BindFile(origin, ttl, records)

    def to_file(self, file_path: str) -> None:
        with open(file_path, "w") as f:
            f.write(str(self))

    def __str__(self) -> str:
        bind = f"$ORIGIN {self.origin}\n"

        if self.ttl is not None:
            bind += f"$TTL {self.ttl}\n"

        for record in self.records:
            bind += f"{record}\n"
        return bind



from dataclasses import dataclass
from enum import Enum


class URLForwardingType(str, Enum):
    temporary = "temporary"
    permanent = "permanent"


@dataclass
class URLForwarding:
    id: str
    subdomain: str
    location: str
    type: URLForwardingType
    include_path: bool
    wildcard: bool

    @staticmethod
    def from_dict(d):
        return URLForwarding(
            id=d["id"],
            subdomain=d["subdomain"],
            location=d["location"],
            type=URLForwardingType[d["type"]],
            include_path=d["includePath"] == "yes",
            wildcard=d["wildcard"] == "yes",
        )



from dataclasses import dataclass
from enum import Enum
from typing import Optional


class DNSRecordType(str, Enum):
    A = "A"
    AAAA = "AAAA"
    MX = "MX"
    CNAME = "CNAME"
    ALIAS = "ALIAS"
    TXT = "TXT"
    NS = "NS"
    SRV = "SRV"
    TLSA = "TLSA"
    CAA = "CAA"

    def __str__(self):
        return self.value


DNS_RECORDS_WITH_PRIORITY = {DNSRecordType.MX, DNSRecordType.SRV}


@dataclass
class DNSRecord:
    id: str
    name: str
    type: DNSRecordType
    content: str
    ttl: int
    prio: Optional[int]
    notes: str

    @staticmethod
    def from_dict(d):
        # only use prio for supported record types since the API returns it for all records with default value 0
        prio = int(d["prio"]) if d["type"] in DNS_RECORDS_WITH_PRIORITY else None
        return DNSRecord(
            id=d["id"],
            name=d["name"],
            type=DNSRecordType[d["type"]],
            content=d["content"],
            ttl=int(d["ttl"]),
            prio=prio,
            notes=d["notes"],
        )


class DNSRestoreMode(Enum):
    clear = 0
    replace = 1
    keep = 2

    def __str__(self):
        return self.name

    @staticmethod
    def from_string(a):
        try:
            return DNSRestoreMode[a]
        except KeyError:
            return a



import json
import tempfile
import unittest
from pathlib import Path
from urllib.parse import urljoin

import responses
from responses import matchers
from responses.registries import OrderedRegistry

from pkb_client.client import (
    API_ENDPOINT,
    DNSRestoreMode,
    PKBClient,
    PKBClientException,
    SSLCertBundle,
)
from pkb_client.client.dns import DNSRecord, DNSRecordType
from pkb_client.client.dnssec import DNSSECRecord
from pkb_client.client.forwarding import URLForwarding, URLForwardingType


class TestClientAuth(unittest.TestCase):
    @responses.activate
    def test_valid_auth(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "ping"),
            json={"status": "SUCCESS", "yourIp": "127.0.0.1"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        ip_address = pkb_client.ping()
        self.assertEqual("127.0.0.1", ip_address)

    @responses.activate
    def test_invalid_auth(self):
        pkb_client = PKBClient("key" + "s", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "ping"),
            json={"status": "ERROR", "message": "Invalid credentials"},
            status=401,
        )
        with self.assertRaises(PKBClientException):
            pkb_client.ping()

    @responses.activate
    def test_ping(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "ping"),
            json={"status": "SUCCESS", "yourIp": "127.0.0.1"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        ip_address = pkb_client.ping()
        self.assertEqual("127.0.0.1", ip_address)

    @responses.activate(registry=OrderedRegistry)
    def test_create_dns_record(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "123456"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": 3600,
                        "prio": None,
                    }
                )
            ],
        )
        assert "123456" == pkb_client.create_dns_record(
            "example.com", DNSRecordType.A, "127.0.0.1", "sub.example.com", 3600
        )

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "234561"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "sub.example.com",
                        "type": "MX",
                        "content": "127.0.0.1",
                        "ttl": 3600,
                        "prio": 2,
                    }
                )
            ],
        )
        assert "234561" == pkb_client.create_dns_record(
            "example.com", DNSRecordType.MX, "127.0.0.1", "sub.example.com", 3600, 2
        )

    def test_create_dns_record_invalid_prio_record_type(self):
        pkb_client = PKBClient("key", "secret")
        with self.assertRaises(ValueError):
            pkb_client.create_dns_record(
                "example.com", DNSRecordType.A, "127.0.0.1", "sub.example.com", 3600, 2
            )

    @responses.activate(registry=OrderedRegistry)
    def test_update_dns_record(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/edit/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "type": "A",
                        "content": "127.0.0.1",
                        "name": "sub.example.com",
                        "ttl": 3600,
                        "prio": None,
                    }
                )
            ],
        )

        success = pkb_client.update_dns_record(
            "example.com",
            "123456",
            DNSRecordType.A,
            "127.0.0.1",
            "sub.example.com",
            3600,
        )
        self.assertTrue(success)

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/edit/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "type": "MX",
                        "content": "127.0.0.1",
                        "name": "sub.example.com",
                        "ttl": 3600,
                        "prio": 2,
                    }
                )
            ],
        )

        success = pkb_client.update_dns_record(
            "example.com",
            "123456",
            DNSRecordType.MX,
            "127.0.0.1",
            "sub.example.com",
            3600,
            2,
        )
        self.assertTrue(success)

    def test_update_dns_records_invalid_prio_record_type(self):
        pkb_client = PKBClient("key", "secret")
        with self.assertRaises(ValueError):
            pkb_client.update_dns_record(
                "example.com",
                "123456",
                DNSRecordType.A,
                "127.0.0.1",
                "sub.example.com",
                3600,
                2,
            )

    @responses.activate
    def test_update_all_dns_records(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/editByNameType/example.com/A/sub"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": 1234,
                        "prio": None,
                    }
                )
            ],
        )

        success = pkb_client.update_all_dns_records(
            "example.com", DNSRecordType.A, "sub", "127.0.0.1", 1234
        )

        self.assertTrue(success)

    def test_update_all_dns_records_all_invalid_prio_record_type(self):
        pkb_client = PKBClient("key", "secret")

        with self.assertRaises(ValueError):
            pkb_client.update_all_dns_records(
                "example.com", DNSRecordType.A, "sub", "127.0.0.1", 1234, 2
            )

    @responses.activate
    def test_delete_dns_record(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/delete/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        success = pkb_client.delete_dns_record("example.com", "123456")

        self.assertTrue(success)

    @responses.activate
    def test_delete_all_dns_records(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/deleteByNameType/example.com/A/sub"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        success = pkb_client.delete_all_dns_records(
            "example.com", DNSRecordType.A, "sub"
        )

        self.assertTrue(success)

    @responses.activate
    def test_get_dns_records(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "123456",
                        "name": "example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": "600",
                        "prio": None,
                        "notes": "",
                    },
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": 600,
                        "prio": None,
                        "notes": "",
                    },
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        records = pkb_client.get_dns_records("example.com")

        expected_records = [
            DNSRecord(
                "123456", "example.com", DNSRecordType.A, "127.0.0.1", 600, None, ""
            ),
            DNSRecord(
                "1234567",
                "sub.example.com",
                DNSRecordType.A,
                "127.0.0.2",
                600,
                None,
                "",
            ),
        ]
        self.assertEqual(expected_records, records)

    @responses.activate
    def test_get_all_dns_records(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieveByNameType/example.com/A/sub"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": 600,
                        "prio": None,
                        "notes": "",
                    }
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        records = pkb_client.get_all_dns_records("example.com", DNSRecordType.A, "sub")

        expected_records = [
            DNSRecord(
                "1234567",
                "sub.example.com",
                DNSRecordType.A,
                "127.0.0.2",
                600,
                None,
                "",
            )
        ]
        self.assertEqual(expected_records, records)

    @responses.activate
    def test_update_dns_servers(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "domain/updateNs/example.com"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "ns": ["ns1.example.com", "ns2.example.com"],
                    }
                )
            ],
        )

        success = pkb_client.update_dns_servers(
            "example.com", ["ns1.example.com", "ns2.example.com"]
        )

        self.assertTrue(success)

    @responses.activate
    def test_get_url_forwards(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "domain/getUrlForwarding/example.com"),
            json={
                "status": "SUCCESS",
                "forwards": [
                    {
                        "id": "123456",
                        "subdomain": "",
                        "location": "https://example.com",
                        "type": "temporary",
                        "includePath": "no",
                        "wildcard": "yes",
                    },
                    {
                        "id": "234567",
                        "subdomain": "sub1",
                        "location": "https://sub1.example.com",
                        "type": "permanent",
                        "includePath": "no",
                        "wildcard": "yes",
                    },
                ],
            },
        )

        forwards = pkb_client.get_url_forwards("example.com")

        expected_forwards = [
            URLForwarding(
                "123456",
                "",
                "https://example.com",
                URLForwardingType.temporary,
                False,
                True,
            ),
            URLForwarding(
                "234567",
                "sub1",
                "https://sub1.example.com",
                URLForwardingType.permanent,
                False,
                True,
            ),
        ]

        self.assertEqual(expected_forwards, forwards)

    @responses.activate
    def test_create_url_forward(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "domain/addUrlForward/example.com"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "subdomain": "sub.example.com",
                        "location": "https://www.example.com",
                        "type": "permanent",
                        "includePath": False,
                        "wildcard": False,
                    }
                )
            ],
        )

        success = pkb_client.create_url_forward(
            "example.com",
            "sub.example.com",
            "https://www.example.com",
            URLForwardingType.permanent,
            False,
            False,
        )

        self.assertTrue(success)

    @responses.activate
    def test_delete_url_forward(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "domain/deleteUrlForward/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        success = pkb_client.delete_url_forward("example.com", "123456")

        self.assertTrue(success)

    @responses.activate
    def test_get_domain_pricing(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "pricing/get"),
            json={
                "status": "SUCCESS",
                "pricing": {
                    "com": {
                        "registration": "42.42",
                        "renewal": "4.2",
                        "transfer": "42.2",
                        "coupons": [],
                    },
                    "test": {
                        "registration": "4.42",
                        "renewal": "44.2",
                        "transfer": "4.2",
                        "coupons": [],
                    },
                },
            },
        )

        pricing = pkb_client.get_domain_pricing()

        expected_pricing = {
            "com": {
                "registration": "42.42",
                "renewal": "4.2",
                "transfer": "42.2",
                "coupons": [],
            },
            "test": {
                "registration": "4.42",
                "renewal": "44.2",
                "transfer": "4.2",
                "coupons": [],
            },
        }

        self.assertEqual(expected_pricing, pricing)

    @responses.activate
    def test_get_ssl_bundle(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "ssl/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "certificatechain": "----BEGIN CERTIFICATE-----\nabc1-----END CERTIFICATE-----\n\n----BEGIN CERTIFICATE-----\nabc2-----END CERTIFICATE-----\n\n----BEGIN CERTIFICATE-----\nabc3-----END CERTIFICATE-----\n",
                "privatekey": "-----BEGIN PRIVATE KEY-----\nabc4-----END PRIVATE KEY-----\n",
                "publickey": "-----BEGIN PUBLIC KEY-----\nabc5-----END PUBLIC KEY-----\n",
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        ssl_cert_bundle = pkb_client.get_ssl_bundle("example.com")

        expected_ssl_cert_bundle = SSLCertBundle(
            certificate_chain="----BEGIN CERTIFICATE-----\nabc1-----END CERTIFICATE-----\n\n----BEGIN CERTIFICATE-----\nabc2-----END CERTIFICATE-----\n\n----BEGIN CERTIFICATE-----\nabc3-----END CERTIFICATE-----\n",
            private_key="-----BEGIN PRIVATE KEY-----\nabc4-----END PRIVATE KEY-----\n",
            public_key="-----BEGIN PUBLIC KEY-----\nabc5-----END PUBLIC KEY-----\n",
        )

        self.assertEqual(expected_ssl_cert_bundle, ssl_cert_bundle)

    @responses.activate
    def test_export_dns_records(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "123456",
                        "name": "example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": "600",
                        "prio": None,
                        "notes": "",
                    },
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": "1200",
                        "prio": None,
                        "notes": "This is a comment",
                    },
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        with tempfile.NamedTemporaryFile() as f:
            pkb_client.export_dns_records("example.com", f.name)

            with open(f.name, "r") as f:
                exported_dns_file = json.load(f)

        expected_exported_dns_file = {
            "123456": {
                "id": "123456",
                "name": "example.com",
                "type": "A",
                "content": "127.0.0.1",
                "ttl": 600,
                "prio": None,
                "notes": "",
            },
            "1234567": {
                "id": "1234567",
                "name": "sub.example.com",
                "type": "A",
                "content": "127.0.0.2",
                "ttl": 1200,
                "prio": None,
                "notes": "This is a comment",
            },
        }

        self.assertEqual(expected_exported_dns_file, exported_dns_file)

    @responses.activate(registry=OrderedRegistry, assert_all_requests_are_fired=True)
    def test_import_dns_records_clear(self):
        pkb_client = PKBClient("key", "secret")

        # first all records should be retrieved
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "123456",
                        "name": "example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": "600",
                        "prio": None,
                        "notes": "",
                    },
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": 600,
                        "prio": None,
                        "notes": "",
                    },
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        # then all records should be deleted
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/delete/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/delete/example.com/1234567"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        # then all records should be imported / created
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "123456"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "",
                        "type": "A",
                        "content": "127.0.0.3",
                        "ttl": 600,
                        "prio": None,
                    }
                )
            ],
        )
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "1234567"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "sub",
                        "type": "A",
                        "content": "127.0.0.4",
                        "ttl": 600,
                        "prio": None,
                    }
                )
            ],
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            filename = Path(temp_dir, "records.json")
            with open(filename, "w") as f:
                json.dump(
                    {
                        "123456": {
                            "id": "123456",
                            "name": "example.com",
                            "type": "A",
                            "content": "127.0.0.3",
                            "ttl": 600,
                            "prio": None,
                        },
                        "1234567": {
                            "id": "1234567",
                            "name": "sub.example.com",
                            "type": "A",
                            "content": "127.0.0.4",
                            "ttl": 600,
                            "prio": None,
                        },
                    },
                    f,
                )

            pkb_client.import_dns_records(
                "example.com", str(filename), DNSRestoreMode.clear
            )

    @responses.activate(registry=OrderedRegistry, assert_all_requests_are_fired=True)
    def test_import_dns_records_replace(self):
        pkb_client = PKBClient("key", "secret")

        # first all records should be retrieved
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "123456",
                        "name": "example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": "600",
                        "prio": None,
                        "notes": "",
                    },
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": 600,
                        "prio": None,
                        "notes": "",
                    },
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        # same record should be updated
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/edit/example.com/1234567"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "type": "A",
                        "content": "127.0.0.3",
                        "name": "sub",
                        "ttl": 600,
                        "prio": None,
                    }
                )
            ],
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            filename = Path(temp_dir, "records.json")
            with open(filename, "w") as f:
                json.dump(
                    {
                        "123451": {
                            "id": "123451",
                            "name": "test.example.com",
                            "type": "A",
                            "content": "127.0.0.4",
                            "ttl": 600,
                            "prio": None,
                        },
                        "1234562": {
                            "id": "1234562",
                            "name": "sub.example.com",
                            "type": "A",
                            "content": "127.0.0.3",
                            "ttl": 600,
                            "prio": None,
                        },
                    },
                    f,
                )

            pkb_client.import_dns_records(
                "example.com", str(filename), DNSRestoreMode.replace
            )

    @responses.activate(registry=OrderedRegistry, assert_all_requests_are_fired=True)
    def test_import_dns_records_keep(self):
        pkb_client = PKBClient("key", "secret")

        # first all records should be retrieved
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "123456",
                        "name": "example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": "600",
                        "prio": None,
                        "notes": "",
                    },
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": 600,
                        "prio": None,
                        "notes": "",
                    },
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        # only new records should be created
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "1234562"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "test",
                        "type": "A",
                        "content": "127.0.0.4",
                        "ttl": 600,
                        "prio": None,
                    }
                )
            ],
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            filename = Path(temp_dir, "records.json")
            with open(filename, "w") as f:
                json.dump(
                    {
                        "123451": {
                            "id": "123451",
                            "name": "test.example.com",
                            "type": "A",
                            "content": "127.0.0.4",
                            "ttl": 600,
                            "prio": None,
                        },
                        "1234562": {
                            "id": "1234562",
                            "name": "sub.example.com",
                            "type": "A",
                            "content": "127.0.0.3",
                            "ttl": 600,
                            "prio": None,
                        },
                    },
                    f,
                )

            pkb_client.import_dns_records(
                "example.com", str(filename), DNSRestoreMode.keep
            )

    @responses.activate(registry=OrderedRegistry, assert_all_requests_are_fired=True)
    def test_import_bind_dns_records(self):
        pkb_client = PKBClient("key", "secret")

        # first all records should be retrieved
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/retrieve/example.com"),
            json={
                "status": "SUCCESS",
                "records": [
                    {
                        "id": "123456",
                        "name": "example.com",
                        "type": "A",
                        "content": "127.0.0.1",
                        "ttl": "600",
                        "prio": None,
                        "notes": "",
                    },
                    {
                        "id": "1234567",
                        "name": "sub.example.com",
                        "type": "A",
                        "content": "127.0.0.2",
                        "ttl": 600,
                        "prio": None,
                        "notes": "",
                    },
                ],
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        # then all records should be deleted
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/delete/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/delete/example.com/1234567"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        # then all records should be imported / created
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "123456"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "",
                        "type": "A",
                        "content": "127.0.0.3",
                        "ttl": 600,
                        "prio": None,
                    }
                )
            ],
        )
        responses.post(
            url=urljoin(API_ENDPOINT, "dns/create/example.com"),
            json={"status": "SUCCESS", "id": "1234567"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "name": "sub",
                        "type": "A",
                        "content": "127.0.0.4",
                        "ttl": 600,
                        "prio": None,
                    }
                )
            ],
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            filename = Path(temp_dir, "records.bind")
            with open(filename, "w") as f:
                f.write(
                    (
                        "$ORIGIN example.com.\n"
                        "$TTL 1234\n"
                        "@ IN SOA dns.example.com. dns2.example.com. (100 300 100 6000 600)\n"
                        "example.com. IN 600 A 127.0.0.3\n"
                        "sub.example.com. 600 IN A 127.0.0.4"
                    )
                )

            pkb_client.import_bind_dns_records(filename, DNSRestoreMode.clear)

    @responses.activate
    def test_get_dnssec_records(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/getDnssecRecords/example.com"),
            json={
                "status": "SUCCESS",
                "records": {
                    "12345": {
                        "keyTag": "12345",
                        "alg": "8",
                        "digestType": "1",
                        "digest": "abc123",
                    },
                    "12346": {
                        "keyTag": "12346",
                        "alg": "8",
                        "digestType": "1",
                        "digest": "abc456",
                        "maxSigLife": 3600,
                        "keyDataFlags": 257,
                        "keyDataProtocol": 3,
                        "keyDataAlgo": 8,
                        "keyDataPubKey": "abc789",
                    },
                },
            },
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )
        dnssec_records = pkb_client.get_dnssec_records("example.com")

        self.assertEqual(2, len(dnssec_records))
        self.assertEqual(
            DNSSECRecord(
                key_tag=12345,
                alg=8,
                digest_type=1,
                digest="abc123",
                max_sig_life=None,
                key_data_flags=None,
                key_data_protocol=None,
                key_data_algo=None,
                key_data_pub_key=None,
            ),
            dnssec_records[0],
        )
        self.assertEqual(
            DNSSECRecord(
                key_tag=12346,
                alg=8,
                digest_type=1,
                digest="abc456",
                max_sig_life=3600,
                key_data_flags=257,
                key_data_protocol=3,
                key_data_algo=8,
                key_data_pub_key="abc789",
            ),
            dnssec_records[1],
        )

    @responses.activate
    def test_create_dnssec_record(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/createDnssecRecord/example.com"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "keyTag": 4242,
                        "alg": 12345,
                        "digestType": 8,
                        "digest": "abc123",
                        "maxSigLife": None,
                        "keyDataFlags": None,
                        "keyDataProtocol": None,
                        "keyDataAlgo": None,
                        "keyDataPubKey": None,
                    }
                )
            ],
        )

        success = pkb_client.create_dnssec_record(
            domain="example.com",
            key_tag=4242,
            alg=12345,
            digest_type=8,
            digest="abc123",
        )
        self.assertTrue(success)

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/createDnssecRecord/example2.com"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {
                        "apikey": "key",
                        "secretapikey": "secret",
                        "keyTag": 4242,
                        "alg": 12345,
                        "digestType": 8,
                        "digest": "abc123",
                        "maxSigLife": 42,
                        "keyDataFlags": 41,
                        "keyDataProtocol": 40,
                        "keyDataAlgo": 39,
                        "keyDataPubKey": "abc42",
                    }
                )
            ],
        )

        success = pkb_client.create_dnssec_record(
            domain="example2.com",
            key_tag=4242,
            alg=12345,
            digest_type=8,
            digest="abc123",
            max_sig_life=42,
            key_data_flags=41,
            key_data_protocol=40,
            key_data_algo=39,
            key_data_pub_key="abc42",
        )
        self.assertTrue(success)

    @responses.activate
    def delete_dnssec_record(self):
        pkb_client = PKBClient("key", "secret")

        responses.post(
            url=urljoin(API_ENDPOINT, "dns/deleteDnssecRecord/example.com/123456"),
            json={"status": "SUCCESS"},
            match=[
                matchers.json_params_matcher(
                    {"apikey": "key", "secretapikey": "secret"}
                )
            ],
        )

        success = pkb_client.delete_dnssec_record("example.com", 123456)
        self.assertTrue(success)


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from importlib import resources

from pkb_client.client.bind_file import BindFile, BindRecord, RecordClass
from pkb_client.client.dns import DNSRecordType
from tests import data


class TestBindFileParsing(unittest.TestCase):
    def test_reading_bind_file(self):
        with self.subTest("With default TTL"):
            with resources.open_text(data, "test.bind") as f:
                bind_file = BindFile.from_file(f.name)

            self.assertEqual("test.com.", bind_file.origin)
            self.assertEqual(1234, bind_file.ttl)
            self.assertEqual(5, len(bind_file.records))
            self.assertEqual(
                BindRecord(
                    "test.com.", 600, RecordClass.IN, DNSRecordType.A, "1.2.3.4"
                ),
                bind_file.records[0],
            )
            self.assertEqual(
                BindRecord(
                    "sub.test.com.", 600, RecordClass.IN, DNSRecordType.A, "4.3.2.1"
                ),
                bind_file.records[1],
            )
            self.assertEqual(
                BindRecord(
                    "test.com.",
                    600,
                    RecordClass.IN,
                    DNSRecordType.AAAA,
                    "2001:db8::1",
                    comment="This is a comment",
                ),
                bind_file.records[2],
            )
            self.assertEqual(
                BindRecord(
                    "test.com.", 1234, RecordClass.IN, DNSRecordType.TXT, "pkb-client"
                ),
                bind_file.records[3],
            )
            self.assertEqual(
                BindRecord(
                    "test.com.",
                    600,
                    RecordClass.IN,
                    DNSRecordType.MX,
                    "mail.test.com.",
                    prio=10,
                ),
                bind_file.records[4],
            )

        with self.subTest("Without default TTL"):
            with resources.open_text(data, "test_no_ttl.bind") as f:
                bind_file = BindFile.from_file(f.name)

            self.assertEqual("test.com.", bind_file.origin)
            self.assertEqual(None, bind_file.ttl)
            self.assertEqual(5, len(bind_file.records))
            self.assertEqual(
                BindRecord(
                    "test.com.", 600, RecordClass.IN, DNSRecordType.A, "1.2.3.4"
                ),
                bind_file.records[0],
            )
            self.assertEqual(
                BindRecord(
                    "sub.test.com.", 600, RecordClass.IN, DNSRecordType.A, "4.3.2.1"
                ),
                bind_file.records[1],
            )
            self.assertEqual(
                BindRecord(
                    "test.com.",
                    700,
                    RecordClass.IN,
                    DNSRecordType.AAAA,
                    "2001:db8::1",
                    comment="This is a comment",
                ),
                bind_file.records[2],
            )
            self.assertEqual(
                BindRecord(
                    "test.com.", 700, RecordClass.IN, DNSRecordType.TXT, "pkb-client"
                ),
                bind_file.records[3],
            )
            self.assertEqual(
                BindRecord(
                    "test.com.",
                    600,
                    RecordClass.IN,
                    DNSRecordType.MX,
                    "mail.test.com.",
                    prio=10,
                ),
                bind_file.records[4],
            )

    def test_writing_bind_file(self):
        records = [
            BindRecord("test.com.", 600, RecordClass.IN, DNSRecordType.A, "1.2.3.4"),
            BindRecord(
                "sub.test.com.", 700, RecordClass.IN, DNSRecordType.A, "4.3.2.1"
            ),
            BindRecord(
                "test.com.", 600, RecordClass.IN, DNSRecordType.AAAA, "2001:db8::1"
            ),
            BindRecord(
                "test.com.", 600, RecordClass.IN, DNSRecordType.TXT, "pkb-client"
            ),
            BindRecord(
                "test.com.",
                600,
                RecordClass.IN,
                DNSRecordType.MX,
                "mail.test.com.",
                prio=10,
            ),
        ]
        bind_file = BindFile("test.com.", 1234, records)

        file_content = (
            "$ORIGIN test.com.\n"
            "$TTL 1234\n"
            "test.com. 600 IN A 1.2.3.4\n"
            "sub.test.com. 700 IN A 4.3.2.1\n"
            "test.com. 600 IN AAAA 2001:db8::1\n"
            "test.com. 600 IN TXT pkb-client\n"
            "test.com. 600 IN MX 10 mail.test.com.\n"
        )

        with tempfile.NamedTemporaryFile() as f:
            bind_file.to_file(f.name)
            with open(f.name) as f2:
                self.assertEqual(file_content.strip(), f2.read().strip())


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass
from typing import Optional


@dataclass
class DNSSECRecord:
    key_tag: int  # The key tag is a 16-bit integer that identifies the DNSKEY record
    alg: int  # Indicates the algorithm used to generate the public key
    digest_type: int  # Indicates the type of digest algorithm used
    digest: str  # The digest of the public key
    max_sig_life: Optional[
        int
    ]  # Indicates the amount of time in seconds the signature is valid
    key_data_flags: Optional[
        int
    ]  # Indicates the key type (Zone-signing or Key-signing)
    key_data_protocol: Optional[int]  # Indicates the protocol used for the key
    key_data_algo: Optional[int]  # Indicates the algorithm used for the key
    key_data_pub_key: Optional[str]  # The public key in base64 format

    @staticmethod
    def from_dict(d):
        return DNSSECRecord(
            key_tag=int(d["keyTag"]),
            alg=int(d["alg"]),
            digest_type=int(d["digestType"]),
            digest=d["digest"],
            max_sig_life=int(d["maxSigLife"]) if "maxSigLife" in d else None,
            key_data_flags=int(d["keyDataFlags"]) if "keyDataFlags" in d else None,
            key_data_protocol=int(d["keyDataProtocol"])
            if "keyDataProtocol" in d
            else None,
            key_data_algo=int(d["keyDataAlgo"]) if "keyDataAlgo" in d else None,
            key_data_pub_key=d["keyDataPubKey"] if "keyDataPubKey" in d else None,
        )



import json
import logging
from hashlib import sha256
from pathlib import Path
from typing import List, Optional, Union
from urllib.parse import urljoin

import dns.resolver
import requests

from pkb_client.client import BindFile
from pkb_client.client.dns import (
    DNS_RECORDS_WITH_PRIORITY,
    DNSRecord,
    DNSRecordType,
    DNSRestoreMode,
)
from pkb_client.client.dnssec import DNSSECRecord
from pkb_client.client.domain import DomainInfo
from pkb_client.client.forwarding import URLForwarding, URLForwardingType
from pkb_client.client.ssl_cert import SSLCertBundle

API_ENDPOINT = "https://api.porkbun.com/api/json/v3/"

logger = logging.getLogger("pkb_client")
logging.basicConfig(level=logging.INFO)


class PKBClientException(Exception):
    def __init__(self, status, message):
        super().__init__(f"{status}: {message}")


class PKBClient:
    """
    API client for Porkbun.
    """

    default_ttl: int = 300

    def __init__(
        self,
        api_key: Optional[str] = None,
        secret_api_key: Optional[str] = None,
        api_endpoint: str = API_ENDPOINT,
        debug: bool = False,
    ) -> None:
        """
        Creates a new PKBClient object.

        :param api_key: the API key used for Porkbun API calls
        :param secret_api_key: the API secret used for Porkbun API calls
        :param api_endpoint: the endpoint of the Porkbun API.
        :param debug: boolean to enable debug logging
        """
        self.api_key = api_key
        self.secret_api_key = secret_api_key
        self.api_endpoint = api_endpoint
        self.debug = debug
        if self.debug:
            logger.setLevel(logging.DEBUG)

    def _get_auth_request_json(self) -> dict:
        """
        Get the request json for the authentication of the Porkbun API calls.

        :return: the request json for the authentication of the Porkbun API calls
        """

        if self.api_key is None or self.secret_api_key is None:
            raise ValueError("api_key and secret_api_key must be set")

        return {"apikey": self.api_key, "secretapikey": self.secret_api_key}

    def ping(self) -> str:
        """
        API ping method: get the current public ip address of the requesting system; can also be used for auth checking.
        See https://api.porkbun.com/api/json/v3/documentation#Authentication for more info.

        :return: the current public ip address of the requesting system
        """

        url = urljoin(self.api_endpoint, "ping")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return json.loads(r.text).get("yourIp", None)
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def create_dns_record(
        self,
        domain: str,
        record_type: DNSRecordType,
        content: str,
        name: Optional[str] = None,
        ttl: int = default_ttl,
        prio: Optional[int] = None,
    ) -> str:
        """
        API DNS create method: create a new DNS record for given domain.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Create%20Record for more info.

        :param domain: the domain for which the DNS record should be created
        :param record_type: the type of the new DNS record
        :param content: the content of the new DNS record
        :param name: the subdomain for which the new DNS record entry should apply; the * can be used for a
                     wildcard DNS record; if not used, then a DNS record for the root domain will be created
        :param ttl: the time to live in seconds of the new DNS record; have to be between 300 and 86400
        :param prio: the priority of the new DNS record (only records of type MX and SRV) otherwise None
        :return: the id of the new created DNS record
        """

        if ttl > 86400 or ttl < self.default_ttl:
            raise ValueError(f"ttl must be between {self.default_ttl} and 86400")

        if prio is not None and record_type not in DNS_RECORDS_WITH_PRIORITY:
            raise ValueError(
                f"Priority can only be set for {DNS_RECORDS_WITH_PRIORITY}"
            )

        url = urljoin(self.api_endpoint, f"dns/create/{domain}")
        req_json = {
            **self._get_auth_request_json(),
            "name": name,
            "type": record_type,
            "content": content,
            "ttl": ttl,
            "prio": prio,
        }
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return str(json.loads(r.text).get("id", None))
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def update_dns_record(
        self,
        domain: str,
        record_id: str,
        record_type: DNSRecordType,
        content: str,
        name: Optional[str] = None,
        ttl: int = default_ttl,
        prio: Optional[int] = None,
    ) -> bool:
        """
        API DNS edit method: edit an existing DNS record specified by the id for a given domain.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Edit%20Record for more info.

        :param domain: the domain for which the DNS record should be edited
        :param record_id: the id of the DNS record which should be edited
        :param record_type: the new type of the DNS record
        :param content: the new content of the DNS record
        :param name: the new value of the subdomain for which the DNS record should apply; the * can be used for a
                     wildcard DNS record; if not set, the record will be set for the record domain
        :param ttl: the new time to live in seconds of the DNS record, have to be between 300 and 86400
        :param prio: the priority of the new DNS record (only records of type MX and SRV) otherwise None

        :return: True if the editing was successful
        """

        if ttl > 86400 or ttl < self.default_ttl:
            raise ValueError(f"ttl must be between {self.default_ttl} and 86400")

        if prio is not None and record_type not in DNS_RECORDS_WITH_PRIORITY:
            raise ValueError(
                f"Priority can only be set for {DNS_RECORDS_WITH_PRIORITY}"
            )

        url = urljoin(self.api_endpoint, f"dns/edit/{domain}/{record_id}")
        req_json = {
            **self._get_auth_request_json(),
            "name": name,
            "type": record_type,
            "content": content,
            "ttl": ttl,
            "prio": prio,
        }
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def update_all_dns_records(
        self,
        domain: str,
        record_type: DNSRecordType,
        subdomain: str,
        content: str,
        ttl: int = default_ttl,
        prio: Optional[int] = None,
    ) -> bool:
        """
        API DNS edit method: edit all existing DNS record matching the domain, record type and subdomain.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Edit%20Record%20by%20Domain,%20Subdomain%20and%20Type for more info.

        :param domain: the domain for which the DNS record should be edited
        :param record_type: the type of the DNS record
        :param subdomain: the subdomain of the DNS record can be empty string for root domain
        :param content: the new content of the DNS record
        :param ttl: the new time to live in seconds of the DNS record, have to be between 300 and 86400
        :param prio: the priority of the new DNS record (only records of type MX and SRV) otherwise None

        :return: True if the editing was successful
        """

        if ttl > 86400 or ttl < self.default_ttl:
            raise ValueError(f"ttl must be between {self.default_ttl} and 86400")

        if prio is not None and record_type not in DNS_RECORDS_WITH_PRIORITY:
            raise ValueError(
                f"Priority can only be set for {DNS_RECORDS_WITH_PRIORITY}"
            )

        url = urljoin(
            self.api_endpoint, f"dns/editByNameType/{domain}/{record_type}/{subdomain}"
        )
        req_json = {
            **self._get_auth_request_json(),
            "type": record_type,
            "content": content,
            "ttl": ttl,
            "prio": prio,
        }
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def delete_dns_record(self, domain: str, record_id: str) -> bool:
        """
        API DNS delete method: delete an existing DNS record specified by the id for a given domain.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Delete%20Record for more info.

        :param domain: the domain for which the DNS record should be deleted
        :param record_id: the id of the DNS record which should be deleted

        :return: True if the deletion was successful
        """

        url = urljoin(self.api_endpoint, f"dns/delete/{domain}/{record_id}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def delete_all_dns_records(
        self, domain: str, record_type: DNSRecordType, subdomain: str
    ) -> bool:
        """
        API DNS delete method: delete all existing DNS record matching the domain, record type and subdomain.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Delete%20Records%20by%20Domain,%20Subdomain%20and%20Type for more info.

        :param domain: the domain for which the DNS record should be deleted
        :param record_type: the type of the DNS record
        :param subdomain: the subdomain of the DNS record can be empty string for root domain

        :return: True if the deletion was successful
        """

        url = urljoin(
            self.api_endpoint,
            f"dns/deleteByNameType/{domain}/{record_type}/{subdomain}",
        )
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_dns_records(
        self, domain, record_id: Optional[str] = None
    ) -> List[DNSRecord]:
        """
        API DNS retrieve method: retrieve all DNS records for given domain if no record id is specified.
        Otherwise, retrieve the DNS record of the specified domain with the given record id.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Retrieve%20Records for more info.

        :param domain: the domain for which the DNS records should be retrieved
        :param record_id: the id of the DNS record which should be retrieved

        :return: list of DNSRecords objects
        """

        if record_id is None:
            url = urljoin(self.api_endpoint, f"dns/retrieve/{domain}")
        else:
            url = urljoin(self.api_endpoint, f"dns/retrieve/{domain}/{record_id}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return [
                DNSRecord.from_dict(record)
                for record in json.loads(r.text).get("records", [])
            ]
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_all_dns_records(
        self, domain: str, record_type: DNSRecordType, subdomain: str
    ) -> List[DNSRecord]:
        """
        API DNS retrieve method: retrieve all DNS records matching the domain, record type and subdomain.
        See https://api.porkbun.com/api/json/v3/documentation#DNS%20Retrieve%20Records%20by%20Domain,%20Subdomain%20and%20Type for more info.

        :param domain: the domain for which the DNS records should be retrieved
        :param record_type: the type of the DNS records
        :param subdomain: the subdomain of the DNS records can be empty string for root domain

        :return: list of DNSRecords objects
        """

        url = urljoin(
            self.api_endpoint,
            f"dns/retrieveByNameType/{domain}/{record_type}/{subdomain}",
        )
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return [
                DNSRecord.from_dict(record)
                for record in json.loads(r.text).get("records", [])
            ]
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def export_dns_records(self, domain: str, filepath: Union[Path, str]) -> bool:
        """
        Export all DNS record from the given domain to a json file.
        This method does not represent a Porkbun API method.
        DNS records with all custom fields like notes are exported.

        :param domain: the domain for which the DNS record should be retrieved and saved
        :param filepath: the filepath where to save the exported DNS records

        :return: True if everything went well
        """

        filepath = Path(filepath)

        logger.debug("retrieve current DNS records...")
        dns_records = self.get_dns_records(domain)

        logger.debug("save DNS records to {} ...".format(filepath))
        # merge the single DNS records into one single dict with the record id as key
        dns_records_dict = dict()
        for record in dns_records:
            dns_records_dict[record.id] = record

        if filepath.exists():
            logger.warning("file already exists, overwriting...")

        with open(filepath, "w") as f:
            json.dump(dns_records_dict, f, default=lambda o: o.__dict__, indent=4)

        logger.info("export finished")

        return True

    def export_bind_dns_records(self, domain: str, filepath: Union[Path, str]) -> bool:
        """
        Export all DNS record from the given domain to a BIND file.
        This method does not represent a Porkbun API method.
        Porkbun DNS record notes are exported as comments.

        :param domain: the domain for which the DNS record should be retrieved and saved
        :param filepath: the filepath where to save the exported DNS records

        :return: True if everything went well
        """

        filepath = Path(filepath)

        logger.debug("retrieve current DNS records...")
        dns_records = self.get_dns_records(domain)

        logger.debug("save DNS records to {} ...".format(filepath))
        # merge the single DNS records into one single dict with the record id as key
        dns_records_dict = dict()
        for record in dns_records:
            dns_records_dict[record.id] = record

        if filepath.exists():
            logger.warning("file already exists, overwriting...")

        # domain header
        bind_file_content = f"$ORIGIN {domain}"

        # SOA record
        soa_records = dns.resolver.resolve(domain, "SOA")
        if soa_records:
            soa_record = soa_records[0]
            bind_file_content += f"\n@ IN SOA {soa_record.mname} {soa_record.rname} ({soa_record.serial} {soa_record.refresh} {soa_record.retry} {soa_record.expire} {soa_record.minimum})"

        # records
        for record in dns_records:
            # name 	record class 	ttl 	record type 	record data
            if record.prio:
                record_content = f"{record.prio} {record.content}"
            else:
                record_content = record.content
            bind_file_content += (
                f"\n{record.name} IN {record.ttl} {record.type} {record_content}"
            )

            if record.notes:
                bind_file_content += f" ; {record.notes}"

        with open(filepath, "w") as f:
            f.write(bind_file_content)

        logger.info("export finished")

        return True

    def import_dns_records(
        self, domain: str, filepath: Union[Path, str], restore_mode: DNSRestoreMode
    ) -> bool:
        """
        Restore all DNS records from a json file to the given domain.
        This method does not represent a Porkbun API method.

        :param domain: the domain for which the DNS record should be restored
        :param filepath: the filepath from which the DNS records are to be restored
        :param restore_mode: The restore mode (DNS records are identified by the record type, name and prio if supported):
            clear: remove all existing DNS records and restore all DNS records from the provided file
            replace: replace only existing DNS records with the DNS records from the provided file,
                     but do not create any new DNS records
            keep: keep the existing DNS records and only create new ones for all DNS records from
                  the specified file if they do not exist

        :return: True if everything went well
        """

        filepath = Path(filepath)

        existing_dns_records = self.get_dns_records(domain)

        with open(filepath, "r") as f:
            exported_dns_records_dict = json.load(f)

        if restore_mode is DNSRestoreMode.clear:
            logger.debug("restore mode: clear")

            try:
                # delete all existing DNS records
                for record in existing_dns_records:
                    self.delete_dns_record(domain, record.id)

                # restore all exported records by creating new DNS records
                for _, exported_record in exported_dns_records_dict.items():
                    name = ".".join(exported_record["name"].split(".")[:-2])
                    self.create_dns_record(
                        domain=domain,
                        record_type=exported_record["type"],
                        content=exported_record["content"],
                        name=name,
                        ttl=exported_record["ttl"],
                        prio=exported_record["prio"],
                    )
            except Exception as e:
                logger.error("something went wrong: {}".format(e.__str__()))
                self.__handle_error_backup__(existing_dns_records)
                logger.error("import failed")
                return False
        elif restore_mode is DNSRestoreMode.replace:
            logger.debug("restore mode: replace")

            try:
                existing_dns_record_hashed = {
                    sha256(
                        f"{record.type}{record.name}{record.prio}".encode()
                    ).hexdigest(): record
                    for record in existing_dns_records
                }
                for record in exported_dns_records_dict.values():
                    record_hash = sha256(
                        f"{record['type']}{record['name']}{record['prio']}".encode()
                    ).hexdigest()
                    existing_record = existing_dns_record_hashed.get(record_hash, None)
                    # check if the exported dns record is different to the existing record,
                    # so we can reduce unnecessary api calls
                    if existing_record is not None and (
                        record["content"] != existing_record.content
                        or record["ttl"] != existing_record.ttl
                        or record["prio"] != existing_record.prio
                    ):
                        self.update_dns_record(
                            domain=domain,
                            record_id=existing_record.id,
                            record_type=record["type"],
                            content=record["content"],
                            name=record["name"].replace(f".{domain}", ""),
                            ttl=record["ttl"],
                            prio=record["prio"],
                        )
            except Exception as e:
                logger.error("something went wrong: {}".format(e.__str__()))
                self.__handle_error_backup__(existing_dns_records)
                logger.error("import failed")
                return False
        elif restore_mode is DNSRestoreMode.keep:
            logger.debug("restore mode: keep")

            existing_dns_record_hashed = {
                sha256(
                    f"{record.type}{record.name}{record.prio}".encode()
                ).hexdigest(): record
                for record in existing_dns_records
            }

            try:
                for record in exported_dns_records_dict.values():
                    record_hash = sha256(
                        f"{record['type']}{record['name']}{record['prio']}".encode()
                    ).hexdigest()
                    existing_record = existing_dns_record_hashed.get(record_hash, None)
                    if existing_record is None:
                        self.create_dns_record(
                            domain=domain,
                            record_type=record["type"],
                            content=record["content"],
                            name=record["name"].replace(f".{domain}", ""),
                            ttl=record["ttl"],
                            prio=record["prio"],
                        )
            except Exception as e:
                logger.error("something went wrong: {}".format(e.__str__()))
                self.__handle_error_backup__(existing_dns_records)
                logger.error("import failed")
                return False
        else:
            raise Exception("restore mode not supported")

        logger.info("import successfully completed")

        return True

    def import_bind_dns_records(
        self, filepath: Union[Path, str], restore_mode: DNSRestoreMode
    ) -> bool:
        """
        Restore all DNS records from a BIND file.
        This method does not represent a Porkbun API method.

        :param filepath: the bind filepath from which the DNS records are to be restored
        :param restore_mode: The restore mode:
            clear: remove all existing DNS records and restore all DNS records from the provided file
        :return: True if everything went well
        """

        bind_file = BindFile.from_file(filepath)

        existing_dns_records = self.get_dns_records(bind_file.origin[:-1])

        if restore_mode is DNSRestoreMode.clear:
            logger.debug("restore mode: clear")

            try:
                # delete all existing DNS records
                for record in existing_dns_records:
                    self.delete_dns_record(bind_file.origin[:-1], record.id)

                # restore all records from BIND file by creating new DNS records
                for record in bind_file.records:
                    # extract subdomain from record name
                    subdomain = record.name.replace(bind_file.origin, "")
                    # replace trailing dot
                    subdomain = subdomain[:-1] if subdomain.endswith(".") else subdomain
                    self.create_dns_record(
                        domain=bind_file.origin[:-1],
                        record_type=record.record_type,
                        content=record.data,
                        name=subdomain,
                        ttl=record.ttl,
                        prio=record.prio,
                    )

            except Exception as e:
                logger.error("something went wrong: {}".format(e.__str__()))
                self.__handle_error_backup__(existing_dns_records)
                logger.error("import failed")
                return False
        else:
            raise Exception(f"restore mode '{restore_mode.value}' not supported")

        logger.info("import successfully completed")

        return True

    def update_dns_servers(self, domain: str, name_servers: List[str]) -> bool:
        """
        Update the name servers of the specified domain.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20Update%20Name%20Servers for more info.

        :return: True if everything went well
        """

        url = urljoin(self.api_endpoint, f"domain/updateNs/{domain}")
        req_json = {**self._get_auth_request_json(), "ns": name_servers}
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200 and json.loads(r.text).get("status", None) == "SUCCESS":
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_dns_servers(self, domain: str) -> List[str]:
        """
        Get the name servers for the given domain.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20Get%20Name%20Servers for more info.

        :return: list of name servers
        """

        url = urljoin(self.api_endpoint, f"domain/getNs/{domain}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return json.loads(r.text).get("ns", [])
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_domains(self, start: int = 0) -> List[DomainInfo]:
        """
        Get all domains for the account in chunks of 1000. If you reach the end of all domains, the list will be empty.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20List%20All for more info.

        :param start: the index of the first domain to retrieve

        :return: list of DomainInfo objects
        """

        url = urljoin(self.api_endpoint, "domain/listAll")

        req_json = {**self._get_auth_request_json(), "start": start}
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return [
                DomainInfo.from_dict(domain)
                for domain in json.loads(r.text).get("domains", [])
            ]
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_url_forwards(self, domain: str) -> List[URLForwarding]:
        """
        Get the url forwarding for the given domain.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20Get%20URL%20Forwarding for more info.

        :return: list of URLForwarding objects
        """

        url = urljoin(self.api_endpoint, f"domain/getUrlForwarding/{domain}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return [
                URLForwarding.from_dict(forwarding)
                for forwarding in json.loads(r.text).get("forwards", [])
            ]
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def create_url_forward(
        self,
        domain: str,
        subdomain: str,
        location: str,
        type: URLForwardingType,
        include_path: bool,
        wildcard: bool,
    ) -> bool:
        """
        Add a url forward for the given domain.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20Add%20URL%20Forward for more info.

        :param domain: the domain for which the url forwarding should be added
        :param subdomain: the subdomain for which the url forwarding should be added, can be empty for root domain
        :param location: the location to which the url forwarding should redirect
        :param type: the type of the url forwarding
        :param include_path: if the path should be included in the url forwarding
        :param wildcard: if the url forwarding should also be applied to all subdomains

        :return: True if the forwarding was added successfully
        """

        url = urljoin(self.api_endpoint, f"domain/addUrlForward/{domain}")
        req_json = {
            **self._get_auth_request_json(),
            "subdomain": subdomain,
            "location": location,
            "type": type,
            "includePath": include_path,
            "wildcard": wildcard,
        }
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def delete_url_forward(self, domain: str, id: str) -> bool:
        """
        Delete an url forward for the given domain.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20Delete%20URL%20Forward for more info.

        :param domain: the domain for which the url forwarding should be deleted
        :param id: the id of the url forwarding which should be deleted

        :return: True if the deletion was successful
        """

        url = urljoin(self.api_endpoint, f"domain/deleteUrlForward/{domain}/{id}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_domain_pricing(self) -> dict:
        """
        Get the pricing for all Porkbun domains.
        See https://api.porkbun.com/api/json/v3/documentation#Domain%20Pricing for more info.

        :return: dict with pricing
        """

        url = urljoin(self.api_endpoint, "pricing/get")
        r = requests.post(url=url)

        if r.status_code == 200:
            return json.loads(r.text)["pricing"]
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_ssl_bundle(self, domain) -> SSLCertBundle:
        """
        API SSL bundle retrieve method: retrieve an SSL bundle for the given domain.
        See https://api.porkbun.com/api/json/v3/documentation#SSL%20Retrieve%20Bundle%20by%20Domain for more info.

        :param domain: the domain for which the SSL bundle should be retrieved

        :return: tuple of intermediate certificate, certificate chain, private key, public key
        """

        url = urljoin(self.api_endpoint, f"ssl/retrieve/{domain}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            ssl_bundle = json.loads(r.text)

            return SSLCertBundle(
                certificate_chain=ssl_bundle["certificatechain"],
                private_key=ssl_bundle["privatekey"],
                public_key=ssl_bundle["publickey"],
            )
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def get_dnssec_records(self, domain: str) -> List[DNSSECRecord]:
        """
        API DNSSEC retrieve method: retrieve all DNSSEC records for the given domain.
        See https://porkbun.com/api/json/v3/documentation#DNSSEC%20Get%20Records for more info.

        :param domain: the domain for which the DNSSEC records should be retrieved
        :return: list of :class:`DNSSECRecord` objects
        """

        url = urljoin(self.api_endpoint, f"dns/getDnssecRecords/{domain}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return [
                DNSSECRecord.from_dict(record)
                for record in json.loads(r.text).get("records", {}).values()
            ]
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def create_dnssec_record(
        self,
        domain: str,
        key_tag: int,
        alg: int,
        digest_type: int,
        digest: str,
        max_sig_life: Optional[int] = None,
        key_data_flags: Optional[int] = None,
        key_data_protocol: Optional[int] = None,
        key_data_algo: Optional[int] = None,
        key_data_pub_key: Optional[str] = None,
    ) -> bool:
        """
        API DNSSEC create method: create a new DNSSEC record for the given domain.
        See https://porkbun.com/api/json/v3/documentation#DNSSEC%20Create%20Record for more info.

        :param domain: the domain for which the DNSSEC record should be created
        :param key_tag: the key tag of the DNSSEC record
        :param alg: algorithm of the DNSSEC record
        :param digest_type: digest type of the DNSSEC record
        :param digest: digest of the DNSSEC record
        :param max_sig_life: maximum signature life of the DNSSEC record in seconds
        :param key_data_flags: key data flags of the DNSSEC record
        :param key_data_protocol: key data protocol of the DNSSEC record
        :param key_data_algo: key data algorithm of the DNSSEC record
        :param key_data_pub_key: key data public key of the DNSSEC record

        :return: True if everything went well
        """

        if max_sig_life is not None and max_sig_life < 0:
            raise ValueError("max_sig_life must be greater than 0")

        url = urljoin(self.api_endpoint, f"dns/createDnssecRecord/{domain}")
        req_json = {
            **self._get_auth_request_json(),
            "keyTag": key_tag,
            "alg": alg,
            "digestType": digest_type,
            "digest": digest,
            "maxSigLife": max_sig_life,
            "keyDataFlags": key_data_flags,
            "keyDataProtocol": key_data_protocol,
            "keyDataAlgo": key_data_algo,
            "keyDataPubKey": key_data_pub_key,
        }
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    def delete_dnssec_record(self, domain: str, key_tag: int) -> bool:
        """
        API DNSSEC delete method: delete an existing DNSSEC record for the given domain.
        See https://porkbun.com/api/json/v3/documentation#DNSSEC%20Delete%20Record for more info.

        :param domain: the domain for which the DNSSEC record should be deleted
        :param key_tag: the key tag of the DNSSEC record
        :return: True if everything went well
        """

        url = urljoin(self.api_endpoint, f"dns/deleteDnssecRecord/{domain}/{key_tag}")
        req_json = self._get_auth_request_json()
        r = requests.post(url=url, json=req_json)

        if r.status_code == 200:
            return True
        else:
            response_json = json.loads(r.text)
            raise PKBClientException(
                response_json.get("status", "Unknown status"),
                response_json.get("message", "Unknown message"),
            )

    @staticmethod
    def __handle_error_backup__(dns_records):
        # merge the single DNS records into one single dict with the record id as key
        dns_records_dict = dict()
        for record in dns_records:
            dns_records_dict[record["id"]] = record

        # generate filename with incremental suffix
        base_backup_filename = "pkb_client_dns_records_backup"
        suffix = 0
        backup_file_path = Path("{}_{}.json".format(base_backup_filename, suffix))
        while backup_file_path.exists():
            suffix += 1
            backup_file_path = Path("{}_{}.json".format(base_backup_filename, suffix))

        with open(backup_file_path, "w") as f:
            json.dump(dns_records_dict, f)

        logger.warning(
            "a backup of your existing dns records was saved to {}".format(
                str(backup_file_path)
            )
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys

sys.path.append("..")

project = "pkb_client"
copyright = "2023-2024, infinityofspace"
author = "infinityofspace"
release = "v2.0.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.viewcode", "sphinx.ext.githubpages"]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



import argparse
import dataclasses
import json
import os
import textwrap
from datetime import datetime

from pkb_client.client import PKBClient, API_ENDPOINT
from pkb_client.client.dns import DNSRecordType, DNSRestoreMode
from pkb_client.client.forwarding import URLForwardingType


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


def main():
    parser = argparse.ArgumentParser(
        description="Python client for the Porkbun API",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent("""
                                    License:
                                    MIT - Copyright (c) Marvin Heptner
                                    """),
    )

    parser.add_argument(
        "-k",
        "--key",
        help='The API key used for Porkbun API calls (usually starts with "pk").',
    )
    parser.add_argument(
        "-s",
        "--secret",
        help='The API secret used for Porkbun API calls (usually starts with "sk").',
    )
    parser.add_argument("--debug", help="Enable debug mode.", action="store_true")
    parser.add_argument(
        "--endpoint", help="The API endpoint to use.", default=API_ENDPOINT
    )

    subparsers = parser.add_subparsers(help="Supported API methods")

    parser_ping = subparsers.add_parser("ping", help="Ping the API Endpoint")
    parser_ping.set_defaults(func=PKBClient.ping)

    parser_dns_create = subparsers.add_parser(
        "create-dns-record", help="Create a new DNS record."
    )
    parser_dns_create.set_defaults(func=PKBClient.create_dns_record)
    parser_dns_create.add_argument(
        "domain", help="The domain for which the new DNS record should be created."
    )
    parser_dns_create.add_argument(
        "record_type",
        help="The type of the new DNS record.",
        choices=list(DNSRecordType),
    )
    parser_dns_create.add_argument("content", help="The content of the new DNS record.")
    parser_dns_create.add_argument(
        "--name",
        help="The subdomain for which the new DNS record should be created."
        "The * can be used for a wildcard DNS record."
        "If not used, then a DNS record for the root domain will be created",
        required=False,
    )
    parser_dns_create.add_argument(
        "--ttl", type=int, help="The ttl of the new DNS record.", required=False
    )
    parser_dns_create.add_argument(
        "--prio", type=int, help="The priority of the new DNS record.", required=False
    )

    parser_dns_edit = subparsers.add_parser(
        "update-dns-record", help="Edit an existing DNS record."
    )
    parser_dns_edit.set_defaults(func=PKBClient.update_dns_record)
    parser_dns_edit.add_argument(
        "domain", help="The domain for which the DNS record should be edited."
    )
    parser_dns_edit.add_argument(
        "record_id", help="The id of the DNS record which should be edited."
    )
    parser_dns_edit.add_argument(
        "record_type",
        help="The new type of the DNS record.",
        choices=list(DNSRecordType),
    )
    parser_dns_edit.add_argument("content", help="The new content of the DNS record.")
    parser_dns_edit.add_argument(
        "--name",
        help="The new value of the subdomain for which the DNS record should apply. "
        "The * can be used for a wildcard DNS record. If not set, the record will "
        "be set for the root domain.",
        required=False,
    )
    parser_dns_edit.add_argument(
        "--ttl", type=int, help="The new ttl of the DNS record.", required=False
    )
    parser_dns_edit.add_argument(
        "--prio", type=int, help="The new priority of the DNS record.", required=False
    )

    parser_dns_delete = subparsers.add_parser(
        "delete-dns-records", help="Delete an existing DNS record."
    )
    parser_dns_delete.set_defaults(func=PKBClient.delete_dns_record)
    parser_dns_delete.add_argument(
        "domain", help="The domain for which the DNS record should be deleted."
    )
    parser_dns_delete.add_argument(
        "record_id", help="The id of the DNS record which should be deleted."
    )

    parser_dns_receive = subparsers.add_parser(
        "get-dns-records", help="Get all DNS records."
    )
    parser_dns_receive.set_defaults(func=PKBClient.get_dns_records)
    parser_dns_receive.add_argument(
        "domain", help="The domain for which the DNS record should be retrieved."
    )

    parser_dns_export = subparsers.add_parser(
        "export-dns-records", help="Save all DNS records to a local json file."
    )
    parser_dns_export.set_defaults(func=PKBClient.export_dns_records)
    parser_dns_export.add_argument(
        "domain",
        help="The domain for which the DNS record should be retrieved and saved.",
    )
    parser_dns_export.add_argument(
        "filepath", help="The filepath where to save the exported DNS records."
    )

    parser_dns_export_bind = subparsers.add_parser(
        "export-bind-dns-records", help="Save all DNS records to a local BIND file."
    )
    parser_dns_export_bind.set_defaults(func=PKBClient.export_bind_dns_records)
    parser_dns_export_bind.add_argument(
        "domain",
        help="The domain for which the DNS record should be retrieved and saved.",
    )
    parser_dns_export_bind.add_argument(
        "filepath", help="The filepath where to save the exported DNS records."
    )

    parser_dns_import = subparsers.add_parser(
        "import-dns-records",
        help="Restore all DNS records from a local json file.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser_dns_import.set_defaults(func=PKBClient.import_dns_records)
    parser_dns_import.add_argument(
        "domain", help="The domain for which the DNS record should be restored."
    )
    parser_dns_import.add_argument(
        "filepath", help="The filepath from which the DNS records are to be restored."
    )
    parser_dns_import.add_argument(
        "restore_mode",
        help="""The restore mode (DNS records are identified by the record type, name and prio if supported):
    clear: remove all existing DNS records and restore all DNS records from the provided file
    replace: replace only existing DNS records with the DNS records from the provided file, but do not create any new DNS records
    keep: keep the existing DNS records and only create new ones for all DNS records from the specified file if they do not exist
    """,
        type=DNSRestoreMode.from_string,
        choices=list(DNSRestoreMode),
    )

    parser_dns_import_bind = subparsers.add_parser(
        "import-bind-dns-records",
        help="Restore all DNS records from a local BIND file.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser_dns_import_bind.set_defaults(func=PKBClient.import_bind_dns_records)
    parser_dns_import_bind.add_argument(
        "filepath", help="The filepath from which the DNS records are to be restored."
    )
    parser_dns_import_bind.add_argument(
        "restore_mode",
        help="""The restore mode (DNS records are identified by the record id):
    clear: remove all existing DNS records and restore all DNS records from the provided file
    """,
        type=DNSRestoreMode.from_string,
        choices=[DNSRestoreMode.clear],
    )

    parser_domain_pricing = subparsers.add_parser(
        "get-domain-pricing", help="Get the pricing for Porkbun domains."
    )
    parser_domain_pricing.set_defaults(func=PKBClient.get_domain_pricing)

    parser_ssl_retrieve = subparsers.add_parser(
        "get-ssl-bundle", help="Retrieve an SSL bundle for given domain."
    )
    parser_ssl_retrieve.set_defaults(func=PKBClient.get_ssl_bundle)
    parser_ssl_retrieve.add_argument(
        "domain", help="The domain for which the SSL bundle should be retrieve."
    )

    parser_update_dns_server = subparsers.add_parser(
        "update-dns-servers", help="Update the DNS servers for a domain."
    )
    parser_update_dns_server.set_defaults(func=PKBClient.update_dns_servers)
    parser_update_dns_server.add_argument(
        "domain", help="The domain for which the DNS servers should be set."
    )
    parser_update_dns_server.add_argument(
        "dns_servers", nargs="+", help="The DNS servers to be set."
    )

    parser_get_dns_server = subparsers.add_parser(
        "get-dns-servers", help="Retrieve the DNS servers for a domain."
    )
    parser_get_dns_server.set_defaults(func=PKBClient.get_dns_servers)
    parser_get_dns_server.add_argument(
        "domain", help="The domain for which the DNS servers should be retrieved."
    )

    parser_list_domains = subparsers.add_parser(
        "get-domains", help="List all domains in this account in chunks of 1000."
    )
    parser_list_domains.set_defaults(func=PKBClient.get_domains)
    parser_list_domains.add_argument(
        "--start",
        type=int,
        help="The start index of the list.",
        default=0,
        required=False,
    )

    parser_get_url_forward = subparsers.add_parser(
        "get-url-forwards", help="Retrieve all URL forwards."
    )
    parser_get_url_forward.set_defaults(func=PKBClient.get_url_forwards)
    parser_get_url_forward.add_argument(
        "domain", help="The domain for which the URL forwards should be retrieved."
    )

    parser_add_url_forward = subparsers.add_parser(
        "create-url-forward", help="Create a new URL forward."
    )
    parser_add_url_forward.set_defaults(func=PKBClient.create_url_forward)
    parser_add_url_forward.add_argument(
        "domain", help="The domain for which the new URL forward should be created."
    )
    parser_add_url_forward.add_argument(
        "location",
        help="The location to which the url forwarding should redirect.",
    )
    parser_add_url_forward.add_argument(
        "type", help="The type of the url forwarding.", choices=list(URLForwardingType)
    )
    parser_add_url_forward.add_argument(
        "--subdomain",
        help="The subdomain for which the url forwarding should be added.",
        required=False,
        default="",
    )
    parser_add_url_forward.add_argument(
        "--include-path",
        help="Whether the path should be included in the url forwarding.",
        action="store_true",
        default=False,
    )
    parser_add_url_forward.add_argument(
        "--wildcard",
        help="Whether the url forwarding should be also applied to subdomains.",
        action="store_true",
        default=False,
    )

    parser_delete_url_forward = subparsers.add_parser(
        "delete-url-forward", help="Delete an existing URL forward."
    )
    parser_delete_url_forward.set_defaults(func=PKBClient.delete_url_forward)
    parser_delete_url_forward.add_argument(
        "domain", help="The domain for which the URL forward should be deleted."
    )
    parser_delete_url_forward.add_argument(
        "id", help="The id of the URL forward which should be deleted."
    )

    args = vars(parser.parse_args())

    debug = args.pop("debug", False)

    func = args.pop("func", None)
    if not func:
        raise argparse.ArgumentError(
            None, "No method specified. Please provide a method and try again."
        )

    endpoint = args.pop("endpoint")
    api_key = args.pop("key")
    api_secret = args.pop("secret")

    # call the api methods which do not require authentication
    if func == PKBClient.get_domain_pricing:
        pkb_client = PKBClient(api_endpoint=endpoint, debug=debug)
        ret = func(pkb_client, **args)

        print(json.dumps(ret, cls=CustomJSONEncoder, indent=4))
        exit(0)

    if api_key is None:
        # try to get the api key from the environment variable or fallback to user input
        api_key = os.environ.get("PKB_API_KEY", "")
        if len(api_key.strip()) == 0:
            while True:
                api_key = input(
                    'Please enter your API key you got from Porkbun (usually starts with "pk"): '
                )
                if len(api_key.strip()) == 0:
                    print("The api key can not be empty.")
                else:
                    break

    if api_secret is None:
        # try to get the api secret from the environment variable or fallback to user input
        api_secret = os.environ.get("PKB_API_SECRET", "")
        if len(api_secret.strip()) == 0:
            while True:
                api_secret = input(
                    'Please enter your API key secret you got from Porkbun (usually starts with "sk"): '
                )
                if len(api_secret.strip()) == 0:
                    print("The api key secret can not be empty.")
                else:
                    break

    pkb_client = PKBClient(
        api_key=api_key, secret_api_key=api_secret, api_endpoint=endpoint, debug=debug
    )

    ret = func(pkb_client, **args)

    print(json.dumps(ret, cls=CustomJSONEncoder, indent=4))


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

import pkb_client

with open("Readme.md") as f:
    long_description = f.read()

setup(
    name="pkb_client",
    version=pkb_client.__version__,
    author="infinityofspace",
    url="https://github.com/infinityofspace/pkb_client",
    description="Python client for the Porkbun API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: Name Service (DNS)",
        "Topic :: Utilities",
        "Topic :: System :: Systems Administration",
    ],
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=["setuptools>=39.0.1", "requests>=2.20.0", "dnspython~=2.7"],
    entry_points={
        "console_scripts": [
            "pkb-client = pkb_client.cli.cli:main",
        ]
    },
)



from dataclasses import dataclass


@dataclass
class SSLCertBundle:
    # The complete certificate chain.
    certificate_chain: str

    # The private key.
    private_key: str

    # The public key.
    public_key: str



from dataclasses import dataclass
from datetime import datetime


@dataclass
class DomainInfo:
    domain: str
    status: str
    tld: str
    create_date: datetime
    expire_date: datetime
    security_lock: bool
    whois_privacy: bool
    auto_renew: bool
    not_local: bool

    @staticmethod
    def from_dict(d):
        return DomainInfo(
            domain=d["domain"],
            status=d["status"],
            tld=d["tld"],
            create_date=datetime.fromisoformat(d["createDate"]),
            expire_date=datetime.fromisoformat(d["expireDate"]),
            security_lock=bool(d["securityLock"]),
            whois_privacy=bool(d["whoisPrivacy"]),
            auto_renew=bool(d["autoRenew"]),
            not_local=bool(d["notLocal"]),
        )
